import os
import re
import libtorrent as lt
from .handle import TorrentHandleWrapper
from .constants import TRACKERS, DHT_NODES


class LibTorrentSession:
    def __init__(self):
        super(LibTorrentSession, self).__init__()
        self.session = lt.session()

        settings = self.session.get_settings()
        settings.update(
            {
                "alert_mask": lt.alert.category_t.all_categories,
                "announce_to_all_tiers": True,
                "announce_to_all_trackers": True,
                "upload_rate_limit": 0,
                "connections_limit": 500,
                "enable_outgoing_utp": False,
                "enable_incoming_utp": False,
            }
        )

        self.session.apply_settings(settings) if hasattr(
            self.session, "apply_settings"
        ) else self.session.set_settings(settings)

        for node, port in DHT_NODES:
            self.session.add_dht_router(node, port)
        self.session.start_dht()

    def _get_info_hash(self, magnet):
        """Get info_hash from magnet link or torrent_handle

        Args:
            magnet (str or lt.torrent_handle): magnet link or torrent_handle

        Returns:
            str: info_hash
        """

        if isinstance(magnet, str):
            info_hash = re.findall(r"btih:.*", magnet)[0][5:45].lower()
            assert len(info_hash) == 40, "invalid magnet"
            return info_hash

        elif isinstance(magnet, lt.torrent_handle):
            return str(magnet.status().info_hash)

    def _exist(self, info_hash):
        """Check if torrent exist in session

        Args:
            info_hash (str): info_hash

        Returns:
            bool: True if exist
        """

        return bool(self.get_torrent_handle(info_hash))

    def _get_torrent_handle(self, info_hash):
        """Get torrent_handle from info_hash

        Args:
            info_hash (str): info_hash

        Returns:
            lt.torrent_handle: torrent_handle
        """

        for torrent_handle in self.session.get_torrents():
            if (
                str(torrent_handle.info_hash()).strip().lower()
                == info_hash.strip().lower()
            ):
                return torrent_handle

    def _stop(self, handle):
        """Stop torrent

        Args:
            handle (str or lt.torrent_handle): magnet link or torrent_handle

        Returns:
            bool: True if stopped
        """

        if isinstance(handle, str) and self.exist(handle):
            self.session.remove_torrent(self.get_torrent_handle(handle))
            return True
        elif isinstance(handle, lt.torrent_handle):
            self.session.remove_torrent(handle)
            return True

    def add_magnet_uri(self, magnet, save_path, sequential=False, callback=None):
        """Add magnet link to libtorrent session

        Args:
            magnet (str): magnet link
            save_path (str): download path
            sequential (bool, optional): download sequentially. Defaults to False.

        Returns:
            TorrentHandleWrapper: TorrentHandleWrapper
        """

        info_hash = self._get_info_hash(magnet)
        if os.path.basename(save_path) != info_hash:
            save_path = os.path.realpath(os.path.join(save_path, info_hash))
        os.makedirs(save_path, exist_ok=True)

        handle = lt.add_magnet_uri(self.session, magnet, {"save_path": save_path})
        handle.set_sequential_download(sequential)
        handle.force_reannounce()

        return TorrentHandleWrapper(self, handle, callback)



# python -m libtorrentx -m magnet:?xt=urn:btih:4C9B41D664D7B6B23F0BF39AE185858CBADDA3FF

from libtorrentx import LibTorrentSession
import argparse


def main(args):
    session = LibTorrentSession()
    handle = session.add_magnet_uri(
        args.magnet,
        args.output,
        sequential=args.sequential,
    )

    while True:
        status, props = handle.read(sleep=1)
        if not status:
            continue

        print(f"{props['name']}, {props['download_speed_human']}, {props['progress']}%")

        if props["is_finished"]:
            break


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument("-m", "--magnet", help="torrent magnet link", required=True)
    ap.add_argument("-o", "--output", help="download path", default="./downloads")
    ap.add_argument(
        "-s", "--sequential", help="download sequentially", type=bool, default=False
    )
    args = ap.parse_args()

    main(args)



import time
import threading
import traceback


class TorrentHandleWrapper:
    def __init__(self, session, handle, callback=None):
        self.session = session
        self.handle = handle
        self.callback = callback
        self.info_hash = str(self.handle.status().info_hash)
        self.save_path = self.handle.status().save_path
        self.start_callback()

    def start_callback(self):
        """Start callback thread"""
        self.stop_callback = False
        self.thread = None
        if self.callback:
            self.thread = threading.Thread(target=self.handle_callback)
            # self.thread.daemon = True
            self.thread.start()

    def handle_callback(self):
        """Handle callback, this can be used to update to db in the background"""
        while True:
            status, props = self.read(sleep=1)
            if not status:
                continue

            try:
                self.callback(props)
            except Exception:
                traceback.print_exc()

            if props.is_finished or self.stop_callback:
                break

    def __human_speed(self, speed):
        """Convert speed to human readable format

        Args:
            speed (int): speed in bytes

        Returns:
            str: human readable speed
        """

        if speed > 1024**3:
            return f"{round(speed / 1024 ** 3, 2)} GB/s"
        elif speed > 1024**2:
            return f"{round(speed / 1024 ** 2, 2)} MB/s"
        elif speed > 1024:
            return f"{round(speed / 1024, 2)} KB/s"
        else:
            return f"{round(speed, 2)} B/s"

    def props(self):
        """Get torrent properties

        Returns:
            dict: torrent properties
        """

        try:
            s = self.handle.status()
            name = self.handle.name() or "Unknown"
        except Exception as e:
            print(e)
            return {}

        return dict(
            info_hash=self.info_hash,
            name=name,
            download_speed=s.download_rate + 1,
            download_speed_human=self.__human_speed(s.download_rate),
            downloaded_bytes=s.total_wanted_done,
            is_finished=self.handle.is_finished(),
            is_paused=s.paused,
            num_connections=s.num_connections,
            num_peers=s.num_peers,
            num_seeds=s.num_seeds,
            num_trackers=len(self.handle.trackers()),
            progress=int(s.progress * 100),
            queue_position=s.queue_position,
            total_bytes=s.total_wanted,
            upload_speed=s.upload_rate,
            save_path=s.save_path,
        )

    def read(self, sleep=0):
        """Read torrent properties

        Args:
            sleep (int, optional): sleep time. Defaults to 0.

        Returns:
            tuple: (status, props)
        """

        if sleep:
            time.sleep(sleep)

        props = self.props()
        status = props.get("name", "Unknown") != "Unknown"
        return (status, props)

    def stop(self):
        """Stop torrent download

        Returns:
            bool: True if stopped
        """

        self.stop_callback = True
        self.session._stop(self.info_hash)

    def start(self):
        """Start torrent download

        Returns:
            bool: True if started
        """

        self.start_callback()
        self.handle = self.session.add_magnet_uri(
            f"magnet:?xt=urn:btih:{self.info_hash}", self.save_path
        )
        return self.handle

    def __repr__(self):
        return f"<{self.__class__.__name__} info_hash={self.info_hash}>"



TRACKERS = [
    "udp://tracker.opentrackr.org:1337/announce",
    "udp://open.demonii.com:1337/announce",
    "udp://tracker.rarbg.to:80/announce",
    "udp://tracker.blackunicorn.xyz:6969/announce",
    "udp://tracker.filetracker.pl:6969/announce",
    "udp://tracker.istole.it:80/announce",
    "udp://tracker.tiny-vps.com:6969/announce",
    "udp://9.rarbg.me:2710/announce",
    "udp://tracker.eztv.ch:80/announce",
    "udp://tracker.yify-torrents.com:1337/announce",
    "udp://tracker.torrenty.org:6969/announce",
    "udp://tracker.yify-torrents.com:6969/announce",
    "udp://tracker.filetracker.pl:8082/announce",
    "udp://tracker.zer0day.to:1337/announce",
    "udp://tracker.filetracker.pl:8089/announce",
    "udp://tracker.ettv.be:80/announce",
    "udp://tracker.mg64.net:6881/announce",
    "udp://tracker.coppersurfer.tk:80/announce",
    "udp://tracker.piratepublic.com:1337/announce",
    "udp://9.rarbg.to:2710/announce",
    "udp://tracker.filetracker.pl:1337/announce",
    "udp://tracker.yify-torrents.com:9090/announce",
    "udp://tracker.internetwarriors.net:1337/announce",
    "udp://tracker.mg64.net:2710/announce",
    "udp://tracker.filetracker.pl:80/announce",
    "udp://tracker.open-internet.nl:6969/announce",
    "udp://tracker.coppersurfer.tk:6969/announce",
    "udp://tracker.mg64.net:6969/announce",
    "udp://tracker.mgtracker.org:2710/announce",
    "udp://tracker.openbittorrent.com:80/announce",
    "udp://public.popcorn-tracker.org:6969/announce",
    "udp://tracker.yify-torrents.com:80/announce",
    "udp://tracker.yify-torrents.com:8080/announce",
    "udp://exodus.desync.com:6969/announce",
    "udp://tracker.internet",
    "udp://tracker.grepler.com:6969/announce",
    "udp://tracker.justseed.it:1337/announce",
    "udp://tracker.leechers-paradise.org:6969/announce",
    "udp://tracker.torrent.eu.org:451/announce",
    "udp://tracker.cyberia.is:6969/announce",
    "udp://tracker.limetorrents.com:80/announce",
    "http://tracker.thepiratebay.org/announce",
    "udp://tracker.filetracker.pl:8080/announce",
    "udp://tracker.1337x.org:80/announce",
    "udp://9.rarbg.com:2710/announce",
    "udp://tracker.filetracker.pl:8081/announce",
]


DHT_NODES = [
    ("router.bittorrent.com", 6881),
    ("dht.aelitis.com", 6881),
    ("dht.libtorrent.org", 6881),
    ("router.silotis.us", 6881),
    ("dht.libtorrent.org", 25401),
    ("router.utorrent.com", 6881),
    ("router.bitcomet.com", 6881),
    ("dht.transmissionbt.com", 6881),
    ("dht.bittorrent.com", 6881),
    ("router.bittorrent.com", 6881),
]
